@@ -1,6 +1,7 @@
 package aws
 
 import (
+	"context"
 	"fmt"
 	"strings"
 	"time"
@@ -8,6 +9,7 @@ import (
 	"github.com/aws/aws-sdk-go/aws"
 	"github.com/aws/aws-sdk-go/aws/arn"
 	"github.com/aws/aws-sdk-go/aws/awserr"
+	"github.com/aws/aws-sdk-go/aws/endpoints"
 	"github.com/aws/aws-sdk-go/aws/request"
 	"github.com/aws/aws-sdk-go/aws/session"
 	"github.com/aws/aws-sdk-go/service/ec2"
@@ -22,6 +24,9 @@ import (
 	"github.com/sirupsen/logrus"
 	"k8s.io/apimachinery/pkg/util/wait"
 
+	awssession "github.com/openshift/installer/pkg/asset/installconfig/aws"
+	"github.com/openshift/installer/pkg/destroy/providers"
+	"github.com/openshift/installer/pkg/types"
 	"github.com/openshift/installer/pkg/version"
 )
 
@@ -63,6 +68,27 @@ type ClusterUninstaller struct {
 	Session *session.Session
 }
 
+// New returns an AWS destroyer from ClusterMetadata.
+func New(logger logrus.FieldLogger, metadata *types.ClusterMetadata) (providers.Destroyer, error) {
+	filters := make([]Filter, 0, len(metadata.ClusterPlatformMetadata.AWS.Identifier))
+	for _, filter := range metadata.ClusterPlatformMetadata.AWS.Identifier {
+		filters = append(filters, filter)
+	}
+
+	session, err := awssession.GetSession()
+	if err != nil {
+		return nil, err
+	}
+
+	return &ClusterUninstaller{
+		Filters:   filters,
+		Region:    metadata.ClusterPlatformMetadata.AWS.Region,
+		Logger:    logger,
+		ClusterID: metadata.InfraID,
+		Session:   session,
+	}, nil
+}
+
 func (o *ClusterUninstaller) validate() error {
 	if len(o.Filters) == 0 {
 		return errors.Errorf("you must specify at least one tag filter")
@@ -107,7 +133,6 @@ func (o *ClusterUninstaller) Run() error {
 		tagClientNames[tagClient] = "us-east-1"
 	}
 
-	deleted := map[string]struct{}{}
 	iamClient := iam.New(awsSession)
 	iamRoleSearch := &iamRoleSearch{
 		client:  iamClient,
@@ -120,12 +145,18 @@ func (o *ClusterUninstaller) Run() error {
 		logger:  o.Logger,
 	}
 
+	deleted, err := terminateEC2InstancesByTags(ec2.New(awsSession), iamClient, o.Filters, o.Logger)
+	if err != nil {
+		return err
+	}
+
+	tagClientStack := append([]*resourcegroupstaggingapi.ResourceGroupsTaggingAPI(nil), tagClients...)
 	err = wait.PollImmediateInfinite(
 		time.Second*10,
 		func() (done bool, err error) {
 			var loopError error
 			nextTagClients := tagClients[:0]
-			for _, tagClient := range tagClients {
+			for _, tagClient := range tagClientStack {
 				matched := false
 				for _, filter := range o.Filters {
 					o.Logger.Debugf("search for and delete matching resources by tag in %s matching %#+v", tagClientNames[tagClient], filter)
@@ -150,11 +181,6 @@ func (o *ClusterUninstaller) Run() error {
 										continue
 									}
 
-									if parsed.Service == "ec2" && strings.HasPrefix(parsed.Resource, "network-interface/") {
-										arnLogger.Debug("skip direct tag deletion, deleting the VPC should catch this")
-										continue
-									}
-
 									err = deleteARN(awsSession, parsed, filter, arnLogger)
 									if err != nil {
 										arnLogger.Debug(err)
@@ -182,7 +208,7 @@ func (o *ClusterUninstaller) Run() error {
 					o.Logger.Debugf("no deletions from %s, removing client", tagClientNames[tagClient])
 				}
 			}
-			tagClients = nextTagClients
+			tagClientStack = nextTagClients
 
 			o.Logger.Debug("search for IAM roles")
 			arns, err := iamRoleSearch.arns()
@@ -223,7 +249,7 @@ func (o *ClusterUninstaller) Run() error {
 				}
 			}
 
-			return len(tagClients) == 0 && loopError == nil, nil
+			return len(tagClientStack) == 0 && loopError == nil, nil
 		},
 	)
 	if err != nil {
@@ -235,6 +261,12 @@ func (o *ClusterUninstaller) Run() error {
 		o.Logger.Debug(err)
 		return err
 	}
+
+	err = removeSharedTags(context.TODO(), tagClients, tagClientNames, o.Filters, o.Logger)
+	if err != nil {
+		return err
+	}
+
 	return nil
 }
 
@@ -508,7 +540,7 @@ func deleteEC2(session *session.Session, arn arn.ARN, filter Filter, logger logr
 	case "image":
 		return deleteEC2Image(client, id, filter, logger)
 	case "instance":
-		return deleteEC2Instance(client, iam.New(session), id, logger)
+		return terminateEC2Instance(client, iam.New(session), id, logger)
 	case "internet-gateway":
 		return deleteEC2InternetGateway(client, id, logger)
 	case "natgateway":
@@ -527,6 +559,8 @@ func deleteEC2(session *session.Session, arn arn.ARN, filter Filter, logger logr
 		return deleteEC2Volume(client, id, logger)
 	case "vpc":
 		return deleteEC2VPC(client, elb.New(session), elbv2.New(session), id, logger)
+	case "vpc-peering-connection":
+		return deleteEC2VPCPeeringConnection(client, id, logger)
 	default:
 		return errors.Errorf("unrecognized EC2 resource type %s", resourceType)
 	}
@@ -604,7 +638,7 @@ func deleteEC2ElasticIP(client *ec2.EC2, id string, logger logrus.FieldLogger) e
 	return nil
 }
 
-func deleteEC2Instance(ec2Client *ec2.EC2, iamClient *iam.IAM, id string, logger logrus.FieldLogger) error {
+func terminateEC2Instance(ec2Client *ec2.EC2, iamClient *iam.IAM, id string, logger logrus.FieldLogger) error {
 	response, err := ec2Client.DescribeInstances(&ec2.DescribeInstancesInput{
 		InstanceIds: []*string{aws.String(id)},
 	})
@@ -617,35 +651,121 @@ func deleteEC2Instance(ec2Client *ec2.EC2, iamClient *iam.IAM, id string, logger
 
 	for _, reservation := range response.Reservations {
 		for _, instance := range reservation.Instances {
-			// Skip 'terminated' instances since they take a while to really get cleaned up
-			if *instance.State.Name == "terminated" {
-				continue
-			}
-			if instance.IamInstanceProfile != nil {
-				parsed, err := arn.Parse(*instance.IamInstanceProfile.Arn)
-				if err != nil {
-					return errors.Wrap(err, "parse ARN for IAM instance profile")
-				}
-
-				err = deleteIAMInstanceProfile(iamClient, parsed, logger)
-				if err != nil {
-					return errors.Wrapf(err, "deleting %s", parsed.String())
-				}
-			}
-
-			_, err := ec2Client.TerminateInstances(&ec2.TerminateInstancesInput{
-				InstanceIds: []*string{instance.InstanceId},
-			})
+			err = terminateEC2InstanceByInstance(ec2Client, iamClient, instance, logger)
 			if err != nil {
 				return err
 			}
-
-			logger.Info("Deleted")
 		}
 	}
 	return nil
 }
 
+func terminateEC2InstanceByInstance(ec2Client *ec2.EC2, iamClient *iam.IAM, instance *ec2.Instance, logger logrus.FieldLogger) error {
+	// Skip 'shutting-down' and 'terminated' instances since they take a while to get cleaned up
+	if instance.State == nil || *instance.State.Name == "shutting-down" || *instance.State.Name == "terminated" {
+		return nil
+	}
+
+	if instance.IamInstanceProfile != nil {
+		parsed, err := arn.Parse(*instance.IamInstanceProfile.Arn)
+		if err != nil {
+			return errors.Wrap(err, "parse ARN for IAM instance profile")
+		}
+
+		err = deleteIAMInstanceProfile(iamClient, parsed, logger)
+		if err != nil {
+			return errors.Wrapf(err, "deleting %s", parsed.String())
+		}
+	}
+
+	_, err := ec2Client.TerminateInstances(&ec2.TerminateInstancesInput{
+		InstanceIds: []*string{instance.InstanceId},
+	})
+	if err != nil {
+		return err
+	}
+
+	logger.Info("Terminating")
+	return nil
+}
+
+// terminateEC2InstancesByTags loops until there all instances which
+// match the given tags are terminated.
+func terminateEC2InstancesByTags(ec2Client *ec2.EC2, iamClient *iam.IAM, filters []Filter, logger logrus.FieldLogger) (map[string]struct{}, error) {
+	if ec2Client.Config.Region == nil {
+		return nil, errors.New("EC2 client does not have region configured")
+	}
+
+	partition, ok := endpoints.PartitionForRegion(endpoints.DefaultPartitions(), *ec2Client.Config.Region)
+	if !ok {
+		return nil, errors.Errorf("no partition found for region %q", *ec2Client.Config.Region)
+	}
+
+	terminated := map[string]struct{}{}
+	err := wait.PollImmediateInfinite(
+		time.Second*10,
+		func() (done bool, err error) {
+			var loopError error
+			matched := false
+			for _, filter := range filters {
+				logger.Debugf("search for and delete matching instances by tag matching %#+v", filter)
+				instanceFilters := make([]*ec2.Filter, 0, len(filter))
+				for key, value := range filter {
+					instanceFilters = append(instanceFilters, &ec2.Filter{
+						Name:   aws.String("tag:" + key),
+						Values: []*string{aws.String(value)},
+					})
+				}
+				err = ec2Client.DescribeInstancesPages(
+					&ec2.DescribeInstancesInput{Filters: instanceFilters},
+					func(results *ec2.DescribeInstancesOutput, lastPage bool) bool {
+						for _, reservation := range results.Reservations {
+							if reservation.OwnerId == nil {
+								continue
+							}
+
+							for _, instance := range reservation.Instances {
+								if instance.InstanceId == nil || instance.State == nil {
+									continue
+								}
+
+								instanceLogger := logger.WithField("instance", *instance.InstanceId)
+								if *instance.State.Name == "terminated" {
+									arn := fmt.Sprintf("arn:%s:ec2:%s:%s:instance/%s", partition.ID(), *ec2Client.Config.Region, *reservation.OwnerId, *instance.InstanceId)
+									if _, ok := terminated[arn]; !ok {
+										instanceLogger.Debug("Terminated")
+										terminated[arn] = exists
+									}
+									continue
+								}
+
+								matched = true
+								err := terminateEC2InstanceByInstance(ec2Client, iamClient, instance, instanceLogger)
+								if err != nil {
+									instanceLogger.Debug(err)
+									loopError = errors.Wrapf(err, "terminating %s", *instance.InstanceId)
+									continue
+								}
+							}
+						}
+
+						return !lastPage
+					},
+				)
+				if err != nil {
+					err = errors.Wrap(err, "describe instances")
+					logger.Info(err)
+					matched = true
+					loopError = err
+				}
+			}
+
+			return !matched && loopError == nil, nil
+		},
+	)
+	return terminated, err
+}
+
 // This is a bit of hack. Some objects, like Instance Profiles, can not be tagged in AWS.
 // We "normally" find those objects by their relation to other objects. We have found,
 // however, that people regularly delete all of their instances and roles outside of
@@ -856,31 +976,45 @@ func deleteEC2SecurityGroup(client *ec2.EC2, id string, logger logrus.FieldLogge
 	}
 
 	for _, group := range response.SecurityGroups {
-		if len(group.IpPermissions) > 0 {
-			_, err := client.RevokeSecurityGroupIngress(&ec2.RevokeSecurityGroupIngressInput{
-				GroupId:       group.GroupId,
-				IpPermissions: group.IpPermissions,
-			})
-			if err != nil {
-				return errors.Wrap(err, "revoking ingress permissions")
-			}
-			logger.Debug("Revoked ingress permissions")
-		}
-
-		if len(group.IpPermissionsEgress) > 0 {
-			_, err := client.RevokeSecurityGroupEgress(&ec2.RevokeSecurityGroupEgressInput{
-				GroupId:       group.GroupId,
-				IpPermissions: group.IpPermissionsEgress,
-			})
-			if err != nil {
-				return errors.Wrap(err, "revoking egress permissions")
-			}
-			logger.Debug("Revoked egress permissions")
+		err = deleteEC2SecurityGroupObject(client, group, logger)
+		if err != nil {
+			return err
 		}
 	}
 
-	_, err = client.DeleteSecurityGroup(&ec2.DeleteSecurityGroupInput{
-		GroupId: aws.String(id),
+	return nil
+}
+
+func deleteEC2SecurityGroupObject(client *ec2.EC2, group *ec2.SecurityGroup, logger logrus.FieldLogger) error {
+	if group.GroupName != nil && *group.GroupName == "default" {
+		logger.Debug("Skipping default security group")
+		return nil
+	}
+
+	if len(group.IpPermissions) > 0 {
+		_, err := client.RevokeSecurityGroupIngress(&ec2.RevokeSecurityGroupIngressInput{
+			GroupId:       group.GroupId,
+			IpPermissions: group.IpPermissions,
+		})
+		if err != nil {
+			return errors.Wrap(err, "revoking ingress permissions")
+		}
+		logger.Debug("Revoked ingress permissions")
+	}
+
+	if len(group.IpPermissionsEgress) > 0 {
+		_, err := client.RevokeSecurityGroupEgress(&ec2.RevokeSecurityGroupEgressInput{
+			GroupId:       group.GroupId,
+			IpPermissions: group.IpPermissionsEgress,
+		})
+		if err != nil {
+			return errors.Wrap(err, "revoking egress permissions")
+		}
+		logger.Debug("Revoked egress permissions")
+	}
+
+	_, err := client.DeleteSecurityGroup(&ec2.DeleteSecurityGroupInput{
+		GroupId: group.GroupId,
 	})
 	if err != nil {
 		if err.(awserr.Error).Code() == "InvalidGroup.NotFound" {
@@ -893,12 +1027,47 @@ func deleteEC2SecurityGroup(client *ec2.EC2, id string, logger logrus.FieldLogge
 	return nil
 }
 
+func deleteEC2SecurityGroupsByVPC(client *ec2.EC2, vpc string, failFast bool, logger logrus.FieldLogger) error {
+	var lastError error
+	err := client.DescribeSecurityGroupsPages(
+		&ec2.DescribeSecurityGroupsInput{
+			Filters: []*ec2.Filter{
+				{
+					Name:   aws.String("vpc-id"),
+					Values: []*string{&vpc},
+				},
+			},
+		},
+		func(results *ec2.DescribeSecurityGroupsOutput, lastPage bool) bool {
+			for _, group := range results.SecurityGroups {
+				err := deleteEC2SecurityGroupObject(client, group, logger.WithField("security group", *group.GroupId))
+				if err != nil {
+					if lastError != nil {
+						logger.Debug(err)
+					}
+					lastError = errors.Wrapf(err, "deleting EC2 security group %s", *group.GroupId)
+					if failFast {
+						return false
+					}
+				}
+			}
+
+			return !lastPage
+		},
+	)
+
+	if lastError != nil {
+		return lastError
+	}
+	return err
+}
+
 func deleteEC2Snapshot(client *ec2.EC2, id string, logger logrus.FieldLogger) error {
 	_, err := client.DeleteSnapshot(&ec2.DeleteSnapshotInput{
 		SnapshotId: &id,
 	})
 	if err != nil {
-		if err.(awserr.Error).Code() == "InvalidSnapshotID.NotFound" {
+		if err.(awserr.Error).Code() == "InvalidSnapshot.NotFound" {
 			return nil
 		}
 		return err
@@ -973,6 +1142,40 @@ func deleteEC2Subnet(client *ec2.EC2, id string, logger logrus.FieldLogger) erro
 	return nil
 }
 
+func deleteEC2SubnetsByVPC(client *ec2.EC2, vpc string, failFast bool, logger logrus.FieldLogger) error {
+	// FIXME: port to DescribeSubnetsPages once we bump our vendored AWS package past v1.19.30
+	results, err := client.DescribeSubnets(
+		&ec2.DescribeSubnetsInput{
+			Filters: []*ec2.Filter{
+				{
+					Name:   aws.String("vpc-id"),
+					Values: []*string{&vpc},
+				},
+			},
+		},
+	)
+	if err != nil {
+		return err
+	}
+
+	var lastError error
+	for _, subnet := range results.Subnets {
+		err := deleteEC2Subnet(client, *subnet.SubnetId, logger.WithField("subnet", *subnet.SubnetId))
+		if err != nil {
+			err = errors.Wrapf(err, "deleting EC2 subnet %s", *subnet.SubnetId)
+			if failFast {
+				return err
+			}
+			if lastError != nil {
+				logger.Debug(lastError)
+			}
+			lastError = err
+		}
+	}
+
+	return lastError
+}
+
 func deleteEC2Volume(client *ec2.EC2, id string, logger logrus.FieldLogger) error {
 	_, err := client.DeleteVolume(&ec2.DeleteVolumeInput{
 		VolumeId: aws.String(id),
@@ -1001,13 +1204,18 @@ func deleteEC2VPC(ec2Client *ec2.EC2, elbClient *elb.ELB, elbv2Client *elbv2.ELB
 		return v2lbError
 	}
 
-	for _, helper := range [](func(client *ec2.EC2, vpc string, failFast bool, logger logrus.FieldLogger) error){
-		deleteEC2NATGatewaysByVPC,      // not always tagged
-		deleteEC2NetworkInterfaceByVPC, // not always tagged
-		deleteEC2RouteTablesByVPC,      // not always tagged
-		deleteEC2VPCEndpointsByVPC,     // not taggable
+	for _, child := range []struct {
+		helper   func(client *ec2.EC2, vpc string, failFast bool, logger logrus.FieldLogger) error
+		failFast bool
+	}{
+		{helper: deleteEC2NATGatewaysByVPC, failFast: true},      // not always tagged
+		{helper: deleteEC2NetworkInterfaceByVPC, failFast: true}, // not always tagged
+		{helper: deleteEC2RouteTablesByVPC, failFast: true},      // not always tagged
+		{helper: deleteEC2SecurityGroupsByVPC, failFast: false},  // not always tagged
+		{helper: deleteEC2SubnetsByVPC, failFast: true},          // not always tagged
+		{helper: deleteEC2VPCEndpointsByVPC, failFast: true},     // not taggable
 	} {
-		err := helper(ec2Client, id, true, logger)
+		err := child.helper(ec2Client, id, child.failFast, logger)
 		if err != nil {
 			return err
 		}
@@ -1063,6 +1271,20 @@ func deleteEC2VPCEndpointsByVPC(client *ec2.EC2, vpc string, failFast bool, logg
 	return nil
 }
 
+func deleteEC2VPCPeeringConnection(client *ec2.EC2, id string, logger logrus.FieldLogger) error {
+	_, err := client.DeleteVpcPeeringConnection(&ec2.DeleteVpcPeeringConnectionInput{
+		VpcPeeringConnectionId: &id,
+	})
+	if err != nil {
+		if err.(awserr.Error).Code() == "InvalidVpcPeeringConnectionID.NotFound" {
+			return nil
+		}
+		return errors.Wrapf(err, "cannot delete VPC Peering Connection %s", id)
+	}
+	logger.Info("Deleted")
+	return nil
+}
+
 func deleteElasticLoadBalancing(session *session.Session, arn arn.ARN, logger logrus.FieldLogger) error {
 	resourceType, id, err := splitSlash("resource", arn.Resource)
 	if err != nil {
@@ -1597,7 +1819,7 @@ func deleteS3(session *session.Session, arn arn.ARN, logger logrus.FieldLogger)
 		Bucket: aws.String(arn.Resource),
 	})
 	err := s3manager.NewBatchDeleteWithClient(client).Delete(aws.BackgroundContext(), iter)
-	if err != nil {
+	if err != nil && !isBucketNotFound(err) {
 		return err
 	}
 	logger.Debug("Emptied")
@@ -1605,10 +1827,114 @@ func deleteS3(session *session.Session, arn arn.ARN, logger logrus.FieldLogger)
 	_, err = client.DeleteBucket(&s3.DeleteBucketInput{
 		Bucket: aws.String(arn.Resource),
 	})
-	if err != nil {
+	if err != nil && !isBucketNotFound(err) {
 		return err
 	}
 
 	logger.Info("Deleted")
 	return nil
 }
+
+func isBucketNotFound(err interface{}) bool {
+	switch s3Err := err.(type) {
+	case awserr.Error:
+		if s3Err.Code() == "NoSuchBucket" {
+			return true
+		}
+		origErr := s3Err.OrigErr()
+		if origErr != nil {
+			return isBucketNotFound(origErr)
+		}
+	case s3manager.Error:
+		if s3Err.OrigErr != nil {
+			return isBucketNotFound(s3Err.OrigErr)
+		}
+	case s3manager.Errors:
+		if len(s3Err) == 1 {
+			return isBucketNotFound(s3Err[0])
+		}
+	}
+	return false
+}
+
+func removeSharedTags(ctx context.Context, tagClients []*resourcegroupstaggingapi.ResourceGroupsTaggingAPI, tagClientNames map[*resourcegroupstaggingapi.ResourceGroupsTaggingAPI]string, filters []Filter, logger logrus.FieldLogger) error {
+	for _, filter := range filters {
+		for key, value := range filter {
+			if strings.HasPrefix(key, "kubernetes.io/cluster/") {
+				if value == "owned" {
+					if err := removeSharedTag(ctx, tagClients, tagClientNames, key, logger); err != nil {
+						return err
+					}
+				} else {
+					logger.Warnf("Ignoring non-owned cluster key %s: %s for shared-tag removal", key, value)
+				}
+			}
+		}
+	}
+
+	return nil
+}
+
+func removeSharedTag(ctx context.Context, tagClients []*resourcegroupstaggingapi.ResourceGroupsTaggingAPI, tagClientNames map[*resourcegroupstaggingapi.ResourceGroupsTaggingAPI]string, key string, logger logrus.FieldLogger) error {
+	request := &resourcegroupstaggingapi.UntagResourcesInput{
+		TagKeys: []*string{aws.String(key)},
+	}
+
+	removed := map[string]struct{}{}
+	tagClients = append([]*resourcegroupstaggingapi.ResourceGroupsTaggingAPI(nil), tagClients...)
+	for len(tagClients) > 0 {
+		nextTagClients := tagClients[:0]
+		for _, tagClient := range tagClients {
+			logger.Debugf("Search for and remove tags in %s matching %s: shared", tagClientNames[tagClient], key)
+			arns := []string{}
+			err := tagClient.GetResourcesPages(
+				&resourcegroupstaggingapi.GetResourcesInput{TagFilters: []*resourcegroupstaggingapi.TagFilter{{
+					Key:    aws.String(key),
+					Values: []*string{aws.String("shared")},
+				}}},
+				func(results *resourcegroupstaggingapi.GetResourcesOutput, lastPage bool) bool {
+					for _, resource := range results.ResourceTagMappingList {
+						arn := *resource.ResourceARN
+						if _, ok := removed[arn]; !ok {
+							arns = append(arns, arn)
+						}
+					}
+
+					return !lastPage
+				},
+			)
+			if err != nil {
+				err = errors.Wrap(err, "get tagged resources")
+				logger.Info(err)
+				nextTagClients = append(nextTagClients, tagClient)
+				continue
+			}
+			if len(arns) == 0 {
+				logger.Debugf("No matches in %s for %s: shared, removing client", tagClientNames[tagClient], key)
+				continue
+			}
+			nextTagClients = append(nextTagClients, tagClient)
+
+			for i := 0; i < len(arns); i += 20 {
+				request.ResourceARNList = make([]*string, 0, 20)
+				for j := 0; i+j < len(arns) && j < 20; j++ {
+					request.ResourceARNList = append(request.ResourceARNList, aws.String(arns[i+j]))
+				}
+				_, err = tagClient.UntagResourcesWithContext(ctx, request)
+				if err != nil {
+					err = errors.Wrap(err, "untag shared resources")
+					logger.Info(err)
+					continue
+				}
+				for j := 0; i+j < len(arns) && j < 20; j++ {
+					arn := arns[i+j]
+					logger.WithField("arn", arn).Infof("Removed tag %s: shared", key)
+					removed[arn] = exists
+				}
+			}
+		}
+		tagClients = nextTagClients
+	}
+
+	return nil
+}
